@@ -8,8 +8,10 @@ import gzip
 import warnings
 import pandas as pd
 import plotly.io as pio
+import itertools
 
-from typing import Union, Type
+from bson.objectid import ObjectId
+from typing import Union, Type, List
 from tqdm.auto import tqdm
 from hashlib import md5
 from pathlib import Path
@@ -20,7 +22,6 @@ from base64 import b64encode, b64decode
 from urllib.parse import urlparse
 from pyisemail import is_email
 from collections import defaultdict
-from ratelimit import limits, sleep_and_retry
 from pyisemail.diagnosis import BaseDiagnosis
 from swagger_spec_validator.common import SwaggerValidationError
 from jsonschema.exceptions import ValidationError
@@ -61,6 +62,7 @@ VALID_URLS |= {
 }
 SUPPORTED_FILETYPES = (Gz, Jpeg, Png, Gif, Tiff)
 SUPPORTED_MIMES = [t().mime for t in SUPPORTED_FILETYPES]
+DEFAULT_DOWNLOAD_DIR = Path.home() / "mpcontribs-downloads"
 
 j2h = Json2Html()
 pd.options.plotting.backend = "plotly"
@@ -123,7 +125,7 @@ url_format = SwaggerFormat(
 )
 
 
-def chunks(lst, n=250):
+def chunks(lst, n):
     if isinstance(lst, set):
         lst = list(lst)
     elif not isinstance(lst, list):
@@ -135,6 +137,44 @@ def chunks(lst, n=250):
         yield lst[i:to]
 
 
+# https://stackoverflow.com/a/8991553
+def grouper(n, iterable):
+    it = iter(iterable)
+    while True:
+        chunk = tuple(itertools.islice(it, n))
+        if not chunk:
+            return
+        yield chunk
+
+
+def get_session():
+    s = FuturesSession(max_workers=MAX_WORKERS)
+    s.hooks['response'].append(_response_hook)
+    return s
+
+
+def _response_hook(resp, *args, **kwargs):
+    content_type = resp.headers['content-type']
+    if content_type == "application/json":
+        result = resp.json()
+
+        if "data" in result and isinstance(result["data"], list):
+            resp.result = result
+            resp.count = len(result["data"])
+        elif "count" in result and isinstance(result["count"], int):
+            resp.count = result["count"]
+
+        if "warning" in result:
+            print("WARNING", result["warning"])
+        elif "error" in result and isinstance(result["error"], str):
+            print("ERROR", result["error"][:10000] + "...")
+
+    elif content_type == "application/gzip":
+        resp.result = resp.content
+    else:
+        print("ERROR", resp.content.decode("utf-8"))
+
+
 class FidoClientGlobalHeaders(FidoClient):
     def __init__(self, headers=None):
         super().__init__()
@@ -281,6 +321,32 @@ def load_client(apikey=None, headers=None, host=None):
     )
 
 
+def _run_futures(futures, total: int = 0, timeout: int = -1):
+    """helper to run futures/requests"""
+    start = time.perf_counter()
+    total_set = total > 0
+    total = total if total_set else len(futures)
+    responses = {}
+
+    with tqdm(leave=False, total=total) as pbar:
+        for future in as_completed(futures):
+            if not future.cancelled():
+                response = future.result()
+                cnt = response.count if total_set and hasattr(response, "count") else 1
+                pbar.update(cnt)
+
+                if hasattr(future, "track_id") and hasattr(response, "result"):
+                    responses[future.track_id] = response.result
+
+                elapsed = time.perf_counter() - start
+
+                if timeout > 0 and elapsed > timeout:
+                    for future in futures:
+                        future.cancel()
+
+    return responses
+
+
 class Client(SwaggerClient):
     """client to connect to MPContribs API
 
@@ -295,7 +361,7 @@ class Client(SwaggerClient):
     # of Singleton): Since the __dict__ of any instance can be re-bound, Borg rebinds it
     # in its __init__ to a class-attribute dictionary. Now, any reference or binding of an
     # instance attribute will actually affect all instances equally.
-    # TODO ratelimit support in bravado (wrapped around get_entry..., monkey-patch?)
+    # NOTE bravado future doesn't work with concurrent.futures
 
     _shared_state = {}
 
@@ -322,7 +388,9 @@ class Client(SwaggerClient):
             print("headers set => ignoring apikey!")
 
         self.apikey = apikey
-        self.headers = {"x-api-key": apikey} if apikey else headers
+        self.headers = headers or {}
+        self.headers = {"x-api-key": apikey} if apikey else self.headers
+        self.headers["Content-Type"] = "application/json"
         self.host = host
         ssl = host.endswith(".materialsproject.org") and not host.startswith("localhost.")
         self.protocol = "https" if ssl else "http"
@@ -333,6 +401,8 @@ class Client(SwaggerClient):
 
         if "swagger_spec" not in self.__dict__ or (
             self.swagger_spec.http_client.headers != self.headers
+        ) or (
+            self.swagger_spec.spec_dict["host"] != self.host
         ):
             self._load()
 
@@ -428,6 +498,22 @@ class Client(SwaggerClient):
 
         return True
 
+    def _get_per_page_default_max(self, op: str = "get", resource: str = "contributions") -> int:
+        resource = self.swagger_spec.resources[resource]
+        param_spec = getattr(resource, f"{op}_entries").params["per_page"].param_spec
+        return param_spec["default"], param_spec["maximum"]
+
+    def _get_per_page(
+        self, per_page: int, op: str = "get", resource: str = "contributions"
+    ) -> int:
+        _, per_page_max = self._get_per_page_default_max(op=op, resource=resource)
+        return min(per_page_max, per_page)
+
+    def get_project_names(self) -> List[str]:
+        """Retrieve list of project names."""
+        resp = self.projects.get_entries(_fields=["name"]).result()
+        return [p["name"] for p in resp["data"]]
+
     def get_project(self, name: str) -> Type[Dict]:
         """Retrieve full project entry
 
@@ -463,12 +549,14 @@ class Client(SwaggerClient):
         else:
             tid = tid_or_md5
 
+        op = self.swagger_spec.resources["tables"].get_entries
+        per_page = op.params["data_per_page"].param_spec["maximum"]
         table = {"data": []}
         page, pages = 1, None
 
         while pages is None or page <= pages:
             resp = self.tables.get_entry(
-                pk=tid, _fields=["_all"], data_page=page, data_per_page=1000
+                pk=tid, _fields=["_all"], data_page=page, data_per_page=per_page
             ).result()
             table["data"].extend(resp["data"])
             if pages is None:
@@ -610,8 +698,7 @@ class Client(SwaggerClient):
 
             existing_columns.add(k)
 
-        resp = self.projects.get_entries(_fields=["name"]).result()
-        valid_projects = {p["name"] for p in resp["data"]}
+        valid_projects = self.get_project_names()
 
         if name not in valid_projects:
             return {"error": f"{name} doesn't exist or you don't have access!"}
@@ -672,7 +759,7 @@ class Client(SwaggerClient):
         return self.projects.update_entry(pk=name, project=payload).result()
 
     def delete_contributions(
-        self, name: str, per_page: int = 100, max_workers: int = 5, retry: bool = False
+            self, name: str, per_page: int = 100, retry: bool = False, timeout: int = -1
     ):
         """Remove all contributions for a project
 
@@ -682,22 +769,19 @@ class Client(SwaggerClient):
         Args:
             name (str): name of the project for which to delete contributions
             per_page (int): number of contributions to delete per request
-            max_workers (int): maximum number of parallel requests to send at a time
             retry (bool): if True, retry deletion of failed contributions until done
+            timeout (int): cancel remaining requests if timeout exceeded (in seconds)
         """
         tic = time.perf_counter()
-
-        if max_workers > MAX_WORKERS:
-            max_workers = MAX_WORKERS
-            print(f"max_workers reset to max {MAX_WORKERS}")
-
-        cids = self.get_contributions(name)["ids"]
+        query = dict(project=name)
+        cids = self.get_all_ids(query=query, timeout=timeout).get(name, {}).get("ids", [])
         total = len(cids)
         # reset columns to be save (sometimes not all are reset BUGFIX?)
         self.projects.update_entry(pk=name, project={"columns": []}).result()
+        per_page = self._get_per_page(per_page, op="delete")
 
         if cids:
-            with FuturesSession(max_workers=max_workers) as session:
+            with get_session() as session:
                 while cids:
                     futures = [
                         session.delete(
@@ -709,11 +793,13 @@ class Client(SwaggerClient):
                                 "per_page": per_page,
                             },
                         )
-                        for chunk in chunks(cids, n=per_page)
+                        for chunk in chunks(cids, per_page)
                     ]
 
-                    self._run_futures(futures, total=len(cids))
-                    cids = self.get_contributions(name)["ids"]
+                    _run_futures(futures, total=len(cids), timeout=timeout)
+                    cids = self.get_all_ids(
+                        query=query, timeout=timeout
+                    ).get(name, {}).get("ids", [])
 
                     if not retry:
                         break
@@ -730,78 +816,110 @@ class Client(SwaggerClient):
         else:
             print(f"There aren't any contributions to delete for {name}")
 
-    @sleep_and_retry
-    @limits(calls=175, period=60)
-    def get_unique_identifiers_flag(self, name: str) -> bool:
-        """Retrieve value of `unique_identifiers` flag for a project
+    def get_totals(self, query: dict = None, per_page: int = 1) -> tuple:
+        """Retrieve total count and pages for contributions matching query
+
+        See `client.contributions.get_entries()` for query.
 
         Args:
-            name (str): name of the project
-        """
-        return self.projects.get_entry(
-            pk=name, _fields=["unique_identifiers"]
-        ).result()["unique_identifiers"]
+            query (dict): query to select contributions
+            per_page (int): number of contributions per page to calculate correct #pages
 
-    @sleep_and_retry
-    @limits(calls=175, period=60)
-    def get_total_pages(self, name: str, per_page: int) -> int:
-        """Retrieve total number of pages for contributions in a project
+        Returns:
+            tuple of total counts and pages
+        """
+        query = query or {}
+        skip_keys = {"per_page", "_fields", "format"}
+        query = {k: v for k, v in query.items() if k not in skip_keys}
+        per_page = self._get_per_page(per_page)
+        result = self.contributions.get_entries(
+            per_page=per_page, _fields=["id"], **query
+        ).result()
+        return result["total_count"], result["total_pages"]
+
+    def get_all_ids(
+        self, query: dict, include: List[str] = None, timeout: int = -1
+    ) -> dict:
+        """Retrieve a list of existing contribution and component ObjectIds
 
         Args:
-            name (str): name of the project
-            per_page (int): number of contributions per page
-        """
-        return self.contributions.get_entries(
-            project=name, per_page=per_page, _fields=["id"],
-        ).result()["total_pages"]
+            query (dict): query to select contributions
+            include (list): components to include in response
+            timeout (int): cancel remaining requests if timeout exceeded (in seconds)
 
-    def get_contributions(self, name: str) -> dict:
-        """Retrieve a list of existing contributions and their components for a project
-
-        Args:
-            name (str): name of the project
-        """
-        ret = defaultdict(set)
-        ret["unique_identifiers"] = self.get_unique_identifiers_flag(name)
-        pages = self.get_total_pages(name, 250)
-        id_fields = ["id", "identifier"]
-
-        @sleep_and_retry
-        @limits(calls=175, period=60)
-        def get_future(page):
-            future = session.get(
-                f"{self.url}/contributions/",
-                headers=self.headers,
-                params={
-                    "project": name,
-                    "page": page,
-                    "per_page": 250,
-                    "_fields": ",".join(id_fields + COMPONENTS)
+        Returns:
+            {"<project-name>": {
+                "ids": {<set of contributions IDs>},
+                "identifiers": {<set of contribution identifiers>},
+                "structures": {
+                    "ids": {<set of structure IDs>},
+                    "md5s": {<set of structure md5s>}
                 },
-            )
-            setattr(future, "track_id", page)
-            return future
+                "tables": {
+                    "ids": {<set of tables IDs>},
+                    "md5s": {<set of tables md5s>}
+                },
+                "attachments": {
+                    "ids": {<set of attachments IDs>},
+                    "md5s": {<set of attachments md5s>}
+                },
+            }, ...}
+        """
+        include = include or []
+        components = set(x for x in include if x in COMPONENTS)
+        if include and not components:
+            print(f"`include` must be subset of {COMPONENTS}!")
+            return
+
+        ret = {}
+        _, per_page = self._get_per_page_default_max()
+        query = query or {}
+        [query.pop(k, None) for k in ["page", "per_page", "_fields"]]
+        _, pages = self.get_totals(query=query, per_page=per_page)
+        id_fields = {"project", "id", "identifier"}
+        fields = ",".join(id_fields | components)
+        url = f"{self.url}/contributions/"
+
+        # convert lists in query to comma-separated
+        for k, v in query.items():
+            if isinstance(v, list):
+                query[k] = ",".join(v)
+
+        with get_session() as session:
+
+            def get_future(page):
+                params = {"page": page, "per_page": per_page, "_fields": fields}
+                params.update(query)
+                future = session.get(url, headers=self.headers, params=params)
+                setattr(future, "track_id", page)
+                return future
 
-        with FuturesSession(max_workers=MAX_WORKERS) as session:
-            # bravado future doesn't work with concurrent.futures
             futures = [get_future(page + 1) for page in range(pages)]
 
             while futures:
-                responses = self._run_futures(futures)
+                responses = _run_futures(futures, timeout=timeout)
 
                 for resp in responses.values():
                     for contrib in resp["data"]:
-                        ret["ids"].add(contrib["id"])
-                        ret["identifiers"].add(contrib["identifier"])
+                        project = contrib["project"]
+                        if project not in ret:
+                            ret[project] = {k: set() for k in ["ids", "identifiers"]}
 
-                        for component in COMPONENTS:
-                            md5s = set(d["md5"] for d in contrib[component])
-                            ret[component] |= md5s
+                        ret[project]["ids"].add(contrib["id"])
+                        ret[project]["identifiers"].add(contrib["identifier"])
+
+                        for component in components:
+                            if component in contrib:
+                                if component not in ret[project]:
+                                    ret[project][component] = {"ids": set(), "md5s": set()}
+
+                                for d in contrib[component]:
+                                    for k in ["id", "md5"]:
+                                        ret[project][component][f"{k}s"].add(d[k])
 
                 futures = [
-                    future
-                    for future in futures
-                    if future.track_id not in responses.keys()
+                    future for future in futures
+                    if not future.cancelled() and future.track_id not in responses.keys()
                 ]
 
         return ret
@@ -823,6 +941,7 @@ class Client(SwaggerClient):
         if not valid:
             return {"error": valid}
 
+        _, per_page = self._get_per_page_default_max(op="update")
         query = query or {}
         query["project"] = name
         has_more = True
@@ -830,22 +949,13 @@ class Client(SwaggerClient):
 
         while has_more:
             resp = self.contributions.update_entries(
-                contributions=data, _limit=250, **query
+                contributions=data, per_page=per_page, **query
             ).result()
             has_more = resp["has_more"]
             updated += resp["count"]
 
         print(f"Updated {updated} contributions.")
 
-    def get_number_contributions(self, **query) -> int:
-        """Retrieve total number of contributions for query
-
-        See `client.contributions.get_entries()` for keyword arguments used in query.
-        """
-        return self.contributions.get_entries(
-            _fields=["id"], _limit=1, **query
-        ).result()["total_count"]
-
     def publish(self, name: str, recursive: bool = False) -> dict:
         """Publish a project and optionally its contributions
 
@@ -871,12 +981,11 @@ class Client(SwaggerClient):
 
     def submit_contributions(
         self,
-        contributions: list,
-        skip_dupe_check: bool = False,
+        contributions: List[dict],
         ignore_dupes: bool = False,
         retry: bool = False,
         per_page: int = 100,
-        max_workers: int = 3,
+        timeout: int = -1
     ):
         """Submit a list of contributions
 
@@ -902,28 +1011,21 @@ class Client(SwaggerClient):
 
         Args:
             contributions (list): list of contribution dicts to submit
-            skip_dupe_check (bool): skip check for duplicates of identifiers and components
             ignore_dupes (bool): force duplicate components to be submitted
             retry (bool): keep trying until all contributions successfully submitted
             per_page (int): number of contributions to submit in each chunk/request
-            max_workers (int): number of parallel requests to use to submit chunk
+            timeout (int): cancel remaining requests if timeout exceeded (in seconds)
         """
         if not contributions or not isinstance(contributions, list):
             print("Please provide list of contributions to submit.")
             return
 
-        tic = time.perf_counter()
-
-        if max_workers > MAX_WORKERS:
-            max_workers = MAX_WORKERS
-            print(f"max_workers reset to max {MAX_WORKERS}")
-
         # get existing contributions
-        existing = defaultdict(set)
-        existing["unique_identifiers"] = True
+        tic = time.perf_counter()
         project_names = set()
         collect_ids = []
         require_one_of = {"data"} | set(COMPONENTS)
+        per_page = self._get_per_page(per_page)
 
         for idx, c in enumerate(contributions):
             has_keys = require_one_of & c.keys()
@@ -944,20 +1046,28 @@ class Client(SwaggerClient):
 
         id2project = {}
         if collect_ids:
-            resp = self.contributions.get_entries(
-                id__in=collect_ids, _fields=["id", "project"]
-            ).result()
-            id2project = {c["id"]: c["project"] for c in resp["data"]}
-            project_names |= set(id2project.values())
+            resp = self.get_all_ids(
+                query=dict(id__in=collect_ids), timeout=timeout
+            )
+            project_names |= set(resp.keys())
 
+            for project_name, values in resp.items():
+                for cid in values["ids"]:
+                    id2project[cid] = project_name
+
+        print("get existing contributions ...")
         project_names = list(project_names)
+        unique_identifiers = {}
+        existing = defaultdict(dict, self.get_all_ids(
+            query=dict(project__in=project_names), include=COMPONENTS, timeout=timeout
+        ))
+        resp = self.projects.get_entries(
+            name__in=project_names, _fields=["name", "unique_identifiers"]
+        ).result()
 
-        if not skip_dupe_check:
-            print("get existing contributions ...")
-            existing = {
-                project_name: self.get_contributions(project_name)
-                for project_name in project_names
-            }
+        for project in resp["data"]:
+            project_name = project["name"]
+            unique_identifiers[project_name] = project["unique_identifiers"]
 
         # prepare contributions
         print("prepare contributions ...")
@@ -973,8 +1083,8 @@ class Client(SwaggerClient):
             update = "id" in contrib
             project_name = id2project[contrib["id"]] if update else contrib["project"]
             if (
-                not update and existing[project_name]["unique_identifiers"]
-                and contrib["identifier"] in existing[project_name]["identifiers"]
+                not update and unique_identifiers[project_name]
+                and contrib["identifier"] in existing[project_name].get("identifiers", {})
             ):
                 continue
 
@@ -1079,7 +1189,7 @@ class Client(SwaggerClient):
 
                     dupe = bool(
                         digest in digests[project_name][component] or
-                        digest in existing[project_name][component]
+                        digest in existing[project_name].get(component, {}).get("md5s", set())
                     )
 
                     if not ignore_dupes and dupe:
@@ -1096,18 +1206,13 @@ class Client(SwaggerClient):
         # submit contributions
         if contribs:
             print("submit contributions ...")
-            with FuturesSession(max_workers=max_workers) as session:
-                # bravado future doesn't work with concurrent.futures
+            with get_session() as session:
                 total = 0
-                headers = {"Content-Type": "application/json"}
-                headers.update(self.headers)
 
-                @sleep_and_retry
-                @limits(calls=175, period=60)
                 def post_future(chunk):
                     return session.post(
                         f"{self.url}/contributions/",
-                        headers=headers,
+                        headers=self.headers,
                         data=json.dumps(chunk).encode("utf-8"),
                     )
 
@@ -1115,7 +1220,7 @@ class Client(SwaggerClient):
                     pk = cdct.pop("id")
                     return session.put(
                         f"{self.url}/contributions/{pk}/",
-                        headers=headers,
+                        headers=self.headers,
                         data=json.dumps(cdct).encode("utf-8"),
                     )
 
@@ -1126,7 +1231,7 @@ class Client(SwaggerClient):
 
                     while contribs[project_name]:
                         futures = []
-                        for chunk in chunks(contribs[project_name], n=per_page):
+                        for chunk in chunks(contribs[project_name], per_page):
                             post_chunk = []
                             for c in chunk:
                                 if "id" in c:
@@ -1137,23 +1242,29 @@ class Client(SwaggerClient):
                             if post_chunk:
                                 futures.append(post_future(post_chunk))
 
-                        self._run_futures(futures, total=ncontribs)
+                        _run_futures(futures, total=ncontribs, timeout=timeout)
 
-                        if existing[project_name]["unique_identifiers"] and retry:
-                            existing[project_name] = self.get_contributions(project_name)
+                        if unique_identifiers[project_name] and retry:
+                            existing[project_name] = self.get_all_ids(
+                                query=dict(project=project_name), include=COMPONENTS,
+                                timeout=timeout
+                            ).get(project_name, {"identifiers": set()})
+                            unique_identifiers[project_name] = self.projects.get_entry(
+                                pk=project_name, _fields=["unique_identifiers"]
+                            ).result()["unique_identifiers"]
                             contribs[project_name] = [
                                 c for c in contribs[project_name]
                                 if c["identifier"] not in existing[project_name]["identifiers"]
                             ]
                         else:
                             contribs[project_name] = []  # abort retrying
-                            if not existing[project_name]["unique_identifiers"] and retry:
+                            if not unique_identifiers[project_name] and retry:
                                 print("Please resubmit failed contributions manually.")
 
                 end = datetime.utcnow()
-                updated_total = self.get_number_contributions(
+                updated_total, _ = self.get_totals(query=dict(
                     project__in=project_names, last_modified__gt=start, last_modified__lt=end
-                )
+                ))
                 toc = time.perf_counter()
                 dt = (toc - tic) / 60
                 self._load()
@@ -1161,95 +1272,223 @@ class Client(SwaggerClient):
         else:
             print("Nothing to submit.")
 
-    def download_contribution(
+    def download_contributions(
         self,
-        cid: str,
-        outdir: Union[str, Path] = "mpcontribs-downloads",
+        query: dict = None,
+        outdir: Union[str, Path] = DEFAULT_DOWNLOAD_DIR,
         overwrite: bool = False,
-        include: list = []
-    ) -> Path:
-        """Download a single contribution as .json.gz file(s)
+        include: List[str] = None,
+        timeout: int = -1
+    ) -> int:
+        """Download a list of contributions as .json.gz file(s)
 
         Args:
-            cid: contribution ObjectId
-            outdir: optional existing output directory
-            overwrite: force re-download of existing contribution/components
-            include: components to include in download
+            query: query to select contributions
+            outdir: optional output directory
+            overwrite: force re-download
+            include: components to include in downloads
+            timeout: cancel remaining requests if timeout exceeded (in seconds)
+
+        Returns:
+            Number of new downloads written to disk.
         """
+        start = time.perf_counter()
+        query = query or {}
+        include = include or []
         outdir = Path(outdir) or Path(".")
         outdir.mkdir(parents=True, exist_ok=True)
-        if any(x for x in include if x not in COMPONENTS):
+        components = set(x for x in include if x in COMPONENTS)
+        if include and not components:
             print(f"`include` must be subset of {COMPONENTS}!")
             return
 
-        include = set(x for x in include if x in COMPONENTS)
-        fields = ["project"] + list(include)
-        contrib = self.contributions.get_entry(pk=cid, _fields=fields).result()
-        components = []
+        all_ids = self.get_all_ids(query=query, include=components, timeout=timeout)
+        fmt = query.get("format", "json")
+        ndownloads = 0
 
-        if include:
-            proj = self.projects.get_entry(pk=contrib["project"], _fields=["columns"]).result()
-            for column in proj["columns"]:
-                path = column["path"]
-                if path.split(".", 1)[0] in include:
-                    components.append(path)
+        for name, values in all_ids.items():
+            if timeout > 0:
+                timeout -= time.perf_counter() - start
+                if timeout < 1:
+                    return ndownloads
 
-        model = self.get_model("ContributionsSchema")
-        fields = list(k for k in model._properties.keys() if k not in COMPONENTS) + components
+                start = time.perf_counter()
 
-        # download contribution
-        subdir = outdir / cid
-        subdir.mkdir(exist_ok=True)
-        path = subdir / "contribution.json.gz"
+            cids = list(values["ids"])
+            paths, per_page = self._download_resource(
+                resource="contributions", ids=cids, fmt=fmt,
+                outdir=outdir, overwrite=overwrite, timeout=timeout
+            )
+            if paths:
+                npaths = len(paths)
+                ndownloads += npaths
+                print(f"Downloaded {len(cids)} contributions for '{name}' in {npaths} file(s).")
+            else:
+                print(f"No new contributions to download for '{name}'.")
 
-        if not path.exists() or overwrite:
-            content = self.contributions.download_entries(
-                id=cid, short_mime="gz", format="json", _fields=fields
-            ).result()
-            path.write_bytes(content)
-            print(path)
+            for component in components:
+                if timeout > 0:
+                    timeout -= time.perf_counter() - start
+                    if timeout < 1:
+                        return ndownloads
 
-        # download components
-        for component in components:
-            path = subdir / f"{component}.json.gz"
-            if not path.exists() or overwrite:
-                ids = [x["id"] for x in contrib[component]]
-                model = self.get_model(f"{component.capitalize()}Schema")
-                fields = list(model._properties.keys())
-                content = getattr(self, component).download_entries(
-                    id__in=ids, short_mime="gz", format="json", _fields=fields
-                ).result()
-                path.write_bytes(content)
-                print(path)
+                    start = time.perf_counter()
 
-        return subdir
+                ids = list(values[component]["ids"])
+                paths, per_page = self._download_resource(
+                    resource=component, ids=ids, fmt=fmt,
+                    outdir=outdir, overwrite=overwrite, timeout=timeout
+                )
+                if paths:
+                    npaths = len(paths)
+                    ndownloads += npaths
+                    print(f"Downloaded {len(ids)} {component} for '{name}' in {npaths} file(s).")
+                else:
+                    print(f"No new {component} to download for '{name}'.")
 
-    def _run_futures(self, futures, total=None):
-        """helper to run futures/requests"""
-        responses = {}
+        return ndownloads
 
-        with tqdm(leave=False, total=total if total else len(futures)) as pbar:
-            for future in as_completed(futures):
-                response = future.result()
-                status = response.status_code
+    def download_structures(
+        self,
+        ids: List[str],
+        outdir: Union[str, Path] = DEFAULT_DOWNLOAD_DIR,
+        overwrite: bool = False,
+        timeout: int = -1,
+        fmt: str = "json"
+    ) -> Path:
+        """Download a list of structures as a .json.gz file
 
-                if status in {200, 201, 400, 401, 404, 500, 502}:
-                    resp = response.json()
-                    if status in {200, 201}:
-                        if total and "data" in resp:
-                            cnt = len(resp["data"])
-                        elif total and "count" in resp:
-                            cnt = resp["count"]
-                        else:
-                            cnt = 1
-                        pbar.update(cnt)
-                        if hasattr(future, "track_id"):
-                            responses[future.track_id] = resp
-                        if "warning" in resp:
-                            print(resp["warning"])
-                    elif status != 502:
-                        print(resp["error"][:10000] + "...")
-                elif status not in {503, 504}:
-                    print("ERROR", response.content.decode("utf-8"))
+        Args:
+            ids: list of structure ObjectIds
+            outdir: optional output directory
+            overwrite: force re-download
+            timeout: cancel remaining requests if timeout exceeded (in seconds)
+            fmt: download format - "json" or "csv"
 
-        return responses
+        Returns:
+            paths of output files
+        """
+        return self._download_resource(
+            resource="structures", ids=ids, fmt=fmt,
+            outdir=outdir, overwrite=overwrite, timeout=timeout
+        )
+
+    def download_tables(
+        self,
+        ids: List[str],
+        outdir: Union[str, Path] = DEFAULT_DOWNLOAD_DIR,
+        overwrite: bool = False,
+        timeout: int = -1,
+        fmt: str = "json"
+    ) -> Path:
+        """Download a list of tables as a .json.gz file
+
+        Args:
+            ids: list of table ObjectIds
+            outdir: optional output directory
+            overwrite: force re-download
+            timeout: cancel remaining requests if timeout exceeded (in seconds)
+            fmt: download format - "json" or "csv"
+
+        Returns:
+            paths of output files
+        """
+        return self._download_resource(
+            resource="tables", ids=ids, fmt=fmt,
+            outdir=outdir, overwrite=overwrite, timeout=timeout
+        )
+
+    def download_attachments(
+        self,
+        ids: List[str],
+        outdir: Union[str, Path] = DEFAULT_DOWNLOAD_DIR,
+        overwrite: bool = False,
+        timeout: int = -1,
+        fmt: str = "json"
+    ) -> Path:
+        """Download a list of attachments as a .json.gz file
+
+        Args:
+            ids: list of attachment ObjectIds
+            outdir: optional output directory
+            overwrite: force re-download
+            timeout: cancel remaining requests if timeout exceeded (in seconds)
+            fmt: download format - "json" or "csv"
+
+        Returns:
+            paths of output files
+        """
+        return self._download_resource(
+            resource="attachments", ids=ids, fmt=fmt,
+            outdir=outdir, overwrite=overwrite, timeout=timeout
+        )
+
+    def _download_resource(
+        self,
+        resource: str,
+        ids: List[str],
+        outdir: Union[str, Path] = DEFAULT_DOWNLOAD_DIR,
+        overwrite: bool = False,
+        timeout: int = -1,
+        fmt: str = "json"
+    ) -> Path:
+        """Helper to download a list of resources as .json.gz file
+
+        Args:
+            resource: type of resource
+            ids: list of resource ObjectIds
+            outdir: optional output directory
+            overwrite: force re-download
+            timeout: cancel remaining requests if timeout exceeded (in seconds)
+            fmt: download format - "json" or "csv"
+
+        Returns:
+            tuple (paths of output files, objects per path / per_page)
+        """
+        resources = ["contributions"] + COMPONENTS
+        if resource not in resources:
+            print(f"`resource` must be one of {resources}!")
+            return
+
+        formats = {"json", "csv"}
+        if fmt not in formats:
+            print(f"`fmt` must be one of {formats}!")
+            return
+
+        oids = sorted(ObjectId(i) for i in ids if ObjectId.is_valid(i))
+        outdir = Path(outdir) or Path(".")
+        subdir = outdir / resource
+        subdir.mkdir(parents=True, exist_ok=True)
+        model = self.get_model(f"{resource.capitalize()}Schema")
+        fields = list(model._properties.keys())
+        # NOTE use default per_page to avoid "URL too long" error
+        per_page, _ = self._get_per_page_default_max(op="download", resource=resource)
+        chunked_oids = grouper(per_page, map(str, oids))
+        paths, futures = [], []
+
+        with get_session() as session:
+            for chunk in chunked_oids:
+                digest = get_md5({"ids": chunk})
+                path = subdir / f"{digest}.{fmt}.gz"
+
+                if not path.exists() or overwrite:
+                    future = session.get(
+                        f"{self.url}/{resource}/download/gz/",
+                        headers=self.headers,
+                        params={
+                            "format": fmt,
+                            "id__in": ",".join(chunk),
+                            "_fields": ",".join(fields),
+                        },
+                    )
+                    setattr(future, "track_id", path)
+                    futures.append(future)
+
+            if futures:
+                responses = _run_futures(futures, timeout=timeout)
+
+                for path, resp in responses.items():
+                    path.write_bytes(resp)
+                    paths.append(path)
+
+        return paths, per_page
